/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** velodyneCntrlTaskCodels.c
 **
 ** Codels used by the control task CntrlTask
 **
 **/

/* System headers */
#include <math.h>
#include <stdio.h>

/* Genom headers */
#include <portLib.h>

/* Internal headers */
#include "server/velodyneHeader.h"
#include "velodyneBuffer.h"

/* Public variables */
extern velodyneReadTask task;

/*------------------------------------------------------------------------------
 * SetReadModeParams - Control
 *
 * Description: checks the read mode parameters.
 *
 * Reports: BAD_READ_MODE_PARAMS
 *
 * Returns: OK or ERROR
 */
STATUS
velodyneSetReadModeParamsCntrl(velodyneSaveParams* readModeParams, int* report)
{
  /* Checks the file name. */
  if (strcmp(readModeParams->baseName, "") == 0) {
    fprintf(stderr, "velodyne: empty file name\n");
    *report = S_velodyne_BAD_READ_MODE_PARAMS;
    return ERROR;
  }

  /* Checks the number of the extension. */
  if (readModeParams->num < 0) {
    fprintf(stderr, "velodyne: negative file extension number\n");
    *report = S_velodyne_BAD_READ_MODE_PARAMS;
    return ERROR;
  }
  return OK;
}

/*------------------------------------------------------------------------------
 * SetSensorToMainPlayback - Control
 *
 * Description: checks wether the input angles are in the range [0, 2pi[.
 *
 * Reports: S_velodyne_BAD_SENSOR_TO_MAIN_ANGLES
 *
 * Returns: OK or ERROR
 */
STATUS
velodyneSetSensorToMainPlaybackCntrl(POM_EULER* sensorToMain, int* report)
{
  if (sensorToMain->yaw < 0.0 || sensorToMain->yaw >= 2 * M_PI
      || sensorToMain->pitch < 0.0 || sensorToMain->pitch >= 2 * M_PI
      || sensorToMain->roll < 0.0 || sensorToMain->roll >= 2 * M_PI) {
    fprintf(stderr, "velodyne: bad sensor-to-main angles\n");
    *report = S_velodyne_BAD_SENSOR_TO_MAIN_ANGLES;
    return ERROR;
  }
  return OK;
}

/*------------------------------------------------------------------------
 * velodyneStopAcquisitionCntrl  -  control codel of CONTROL request Stop
 *
 * Description:
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneStopAcquisitionCntrl(int* report)
{
  SDI_F->running = 0;
  pthread_join(task.thr, NULL);
  pthread_mutex_destroy(&task.mtx);
  return OK;
}

/*------------------------------------------------------------------------
 * velodyneSetSaveParamsCntrl  -  control codel of request SetSaveParams
 *
 * Description: checks that the suffix number is positive
 * Report: OK
 *                 S_velodyne_BAD_SAVE_PARAMS
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneSetSaveParamsCntrl(velodyneSaveParams* saveParams, int* report)
{

  if (saveParams->num < 0) {
    fprintf (stderr, "velodyne: bad file suffix number\n");
    *report = S_velodyne_BAD_SAVE_PARAMS;
    return ERROR;
  }
  return OK;
}


/*------------------------------------------------------------------------
 * velodyneOneShotCntrl  -  control codel of request OneShot
 *
 * Description: checks that the provided angles are within correct bounds and
 * weel ordered
 * Report: OK
 *                 S_velodyne_BAD_ONE_SHOT_PARAMS
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneOneShotCntrl(velodyneOneShotParams* oneShotParams, int* report)
{
  if (oneShotParams->firstAngle < -180.0 || oneShotParams->firstAngle > 180) {
    fprintf (stderr, "velodyne: respect -180 < firstAngle < 180 \n");
    *report = S_velodyne_BAD_ONE_SHOT_PARAMS;
    return ERROR;
  }
  if (oneShotParams->lastAngle < -180.0 || oneShotParams->lastAngle > 180) {
    fprintf (stderr, "velodyne: respect -180 < lastAngle < 180 \n");
    *report = S_velodyne_BAD_ONE_SHOT_PARAMS;
    return ERROR;
  }
  return OK;
}

/*------------------------------------------------------------------------
 * velodyneSetTargetParamsCnrtl  -  control codel of request SetTargetParams
 *
 * Description: checks that the provided parameters are within correct bounds
 * Report: OK
 *                 S_velodyne_BAD_TARGET_TRACKING_PARAMETER
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneSetTargetParamsCnrtl (velodyneTargetParams* trackingParams,
			      int* report)
{

  if ((trackingParams->minReflectance < 0) 
      || (trackingParams->maxReflectance > 255)
      || (trackingParams->minReflectance < 0)
      || (trackingParams->maxReflectance > 255)
      || (trackingParams->minReflectance > trackingParams->maxReflectance) ) {
    fprintf (stderr, "### velodyne: respect 0 <= minReflectance <= 255\n"
	     "###\t\tAND 0 <= maxReflectance <= 255\n"
	     "###\t\tAND minReflectance < maxReflectance\n");
    *report = S_velodyne_BAD_TARGET_TRACKING_PARAMETER;
    return ERROR;
  }

  if (trackingParams->maxDistance < 0)  {
    fprintf (stderr, "### velodyne: enter a max target detection distance > 0\n");
    *report = S_velodyne_BAD_TARGET_TRACKING_PARAMETER;
    return ERROR;
  }

  if ((trackingParams->maxWidth <= trackingParams->minWidth)
      || (trackingParams->maxHeight <= trackingParams->minHeight)) {
    fprintf (stderr, "### velodyne: respect 'maxWidth > minWidth' and 'maxHeight > minHeight'\n");
    *report = S_velodyne_BAD_TARGET_TRACKING_PARAMETER;
    return ERROR;
  }

  return OK;
}

/*------------------------------------------------------------------------
 * velodyneStopTargetDetectionCntrl  -  control codel of request StopTargetDetection
 *
 * Returns:    OK
 */

STATUS
velodyneStopTargetDetectionCntrl (int* report)
{
  SDI_F->targetDetectionParams.detectionOn = VELODYNE_FALSE;

  return OK;
}

/*------------------------------------------------------------------------------
 * velodyneSetTargetTrackingParams
 */

/*
 * Ctrl
 * Reports: OK
 *          BAD_TARGET_TRACKING_PARAMETER
 * Returns: OK ERROR
 */
STATUS
velodyneSetTargetTrackingParamsCtrl (
    velodyneTargetTrackingParams* targetTrackingParams, int* report)
{
  if (targetTrackingParams->maxTargetMovement < 0) {
    fprintf (stderr,
             "velodyne: maximum target movement must be higher than 0\n");
    *report = S_velodyne_BAD_TARGET_TRACKING_PARAMETER;
    return ERROR;
  }
  return OK;
}

/*------------------------------------------------------------------------------
 * SetTargetDetectionFileOutput - Ctrl
 *
 * Description: sets the target detection file output parameters.
 *
 * Reports: BAD_TARGET_DETECTION_FILE_OUTPUT_PARAMETER
 *
 * Returns: OK
 *          ERROR
 */
STATUS
velodyneSetTargetDetectionFileOutputCtrl(
    velodyneTargetDetectionFileOutput* targetDetectionFileOutput,
    int* report)
{
  if (targetDetectionFileOutput->enabled == VELODYNE_TRUE
      && strcmp(targetDetectionFileOutput->fileName, "") == 0) {
    fprintf(stderr, "velodyne: file name mustn't be empty\n");
    *report = S_velodyne_BAD_TARGET_DETECTION_FILE_OUTPUT_PARAMETER;
    return ERROR;
  }
  return OK;
}

/*------------------------------------------------------------------------------
 * velodyneSetTargetTrackingFileOutputParams
 */

/*
 * Ctrl
 * Reports: OK
 *          BAD_TARGET_TRACKING_FILE_OUTPUT_PARAMETER
 * Returns: OK ERROR
 */
STATUS
velodyneSetTargetTrackingFileOutputParamsCtrl (
    velodyneTargetTrackingFileOutputParams* targetTrackingFileOutputParams,
    int* report)
{
  if (targetTrackingFileOutputParams->enableFileOutput == VELODYNE_TRUE
      && strcmp(targetTrackingFileOutputParams->outputFileName, "") == 0) {
    fprintf (stderr, "velodyne: output file name mustn't be empty\n");
    *report = S_velodyne_BAD_TARGET_TRACKING_FILE_OUTPUT_PARAMETER;
    return ERROR;
  }
  return OK;
}
