/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyneBuffer.h"

/* System headers */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* External headers */
#include <libimages3d.h>

/* Directives */
#define NUM_CHARS_POS_FILE_HEADER 27
#define NUM_ITEMS_POS_FILE_LINE 7

/* Public variables */
extern VELODYNE_INTERFACE* interface;
extern velodyneReadTask task;
extern POSTER_ID sensorPosPosterId;

void
velodyneCreateBuffer(int numPackets)
{
  task.buffer.nbPackets = numPackets;
  task.buffer.read = 0;
  task.buffer.write = 0;
  task.buffer.dataOverwriting = false;
  task.buffer.packets = malloc (numPackets * sizeof (velodynePacket));
  if (task.buffer.packets == NULL)
    fprintf (stderr, "velodyne: cannot allocate buffer for %d packets\n",
             numPackets);
}

void
velodyneDestroyBuffer()
{
  if (task.buffer.packets != NULL) {
    free (task.buffer.packets);
    task.buffer.packets = NULL;
  }
  task.buffer.nbPackets = 0;
  task.buffer.read = 0;
  task.buffer.write = 0;
  task.buffer.dataOverwriting = false;
}

/*
 * XXX Unlocked access to SDI_F->running, but we can guess it is atomic
 * XXX No way to return a useful error to system as the thread run in
 * background
 */
void*
velodyneReadLive()
{
  static POM_SENSOR_POS pos;
  static int i = 0;

  while (SDI_F->running != 0) {
    pthread_mutex_lock (&task.mtx);
    if (task.needOneShot)
      pthread_cond_wait (&task.cond, &task.mtx);

    if (velodyne_read_packet (interface,
                              &(task.buffer.packets[task.buffer.write].packet))
        != VELODYNE_PAYLOAD_SIZE) {
      fprintf (stderr, "velodyne: cannot read packet\n");
      SDI_F->running = 0;
      pthread_mutex_unlock (&task.mtx);
      return NULL;
    }
    if (SDI_F->initParams.usePom == VELODYNE_TRUE) {
      if (i++ % 50 == 0) {
        if (velodynePOM_SENSOR_POSPosterRead (sensorPosPosterId, &pos)
            == ERROR) {
          fprintf (stderr, "velodyne: unable to read pom poster\n");
          SDI_F->running = 0;
          pthread_mutex_unlock (&task.mtx);
          return NULL;
        }
      }
      task.buffer.packets[task.buffer.write].position = pos;
    }

    task.buffer.write = (task.buffer.write + 1) % task.buffer.nbPackets;
    if (task.buffer.write == task.buffer.read)
      task.buffer.dataOverwriting = true;
    pthread_mutex_unlock (&task.mtx);
  }

  return NULL;
}

velodyneReaderResult
velodynePosFileOpen(const char* fileName, FILE** fileId)
{
  char aux[VELODYNE_MAX_STRLEN];
  FILE* f = fopen(fileName, "r");
  if (f == NULL) {
    *fileId = NULL;
    return VELODYNE_READER_ERR_CANNOT_OPEN_POS_FILE;
  }
  if (fgets(aux, NUM_CHARS_POS_FILE_HEADER, f) == NULL) {
    fclose(f);
    *fileId = NULL;
    return VELODYNE_READER_ERR_BAD_POS_FILE_FORMAT;
  }
  *fileId = f;
  return VELODYNE_READER_OK;
}

velodyneReaderResult
velodynePosFileRead(FILE* fileId, double* timespec, POM_EULER* pos)
{
  int result;
  result = fscanf(fileId, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
                  timespec,
                  &pos->yaw, &pos->pitch, &pos->roll,
                  &pos->x, &pos->y, &pos->z);
  if (result == EOF)
    return VELODYNE_READER_ERR_END_POS_FILE;
  else if (result != NUM_ITEMS_POS_FILE_LINE)
    return VELODYNE_READER_ERR_CANNOT_READ_POS_FILE;
  return VELODYNE_READER_OK;
}

void*
velodyneReadPlayback(void* positionFileId)
{
  FILE* posFileId = (FILE*) positionFileId;
  double tCurr, tNext, tPacket;
  POM_EULER mainToOriginCurr, mainToOriginNext;
  bool oneShotRequired = false;
  double errCurr, errNext;
  velodyneReaderResult posFileReadResult;
  int dateCurr = 0; /* Number of current position. */

  /* Initializes current position. */
  posFileReadResult = velodynePosFileRead(posFileId, &tCurr, &mainToOriginCurr);
  if (posFileReadResult == VELODYNE_READER_ERR_END_POS_FILE) {
    fprintf(stderr, "velodyne: end of position file\n");
    SDI_F->running = 0;
    return NULL;
  } else if (posFileReadResult == VELODYNE_READER_ERR_CANNOT_READ_POS_FILE) {
    fprintf(stderr, "velodyne: cannot read position file\n");
    SDI_F->running = 0;
    return NULL;
  }

  /* Initializes next position. */
  posFileReadResult = velodynePosFileRead(posFileId, &tNext, &mainToOriginNext);
  if (posFileReadResult == VELODYNE_READER_ERR_END_POS_FILE) {
    fprintf(stderr, "velodyne: end of position file\n");
    SDI_F->running = 0;
    return NULL;
  } else if (posFileReadResult == VELODYNE_READER_ERR_CANNOT_READ_POS_FILE) {
    fprintf(stderr, "velodyne: cannot read position file\n");
    SDI_F->running = 0;
    return NULL;
  }

  /* Reading loop. */
  while (SDI_F->running != 0) {
    /* Pre-reading operations. */
    pthread_mutex_lock(&task.mtx);
    if (task.buffer.dataOverwriting)
      pthread_cond_wait(&task.cond, &task.mtx);

    /* Puts Velodyne packet in the buffer. */
    if (velodyne_read_packet(
            interface, &(task.buffer.packets[task.buffer.write].packet))
        != VELODYNE_PAYLOAD_SIZE) {
      if (velodyne_read_packet(
              interface, &(task.buffer.packets[task.buffer.write].packet))
          != VELODYNE_PAYLOAD_SIZE) {
        fprintf(stderr, "velodyne: cannot read packet\n");
        SDI_F->running = 0;
        pthread_mutex_unlock(&task.mtx);
        return NULL;
      }
    }

    /* Computes initial timespec errors for the packet. */
    tPacket
        = task.buffer.packets[task.buffer.write].packet.ts.tv_sec
          + task.buffer.packets[task.buffer.write].packet.ts.tv_usec * 1.0E-6;
    errCurr = fabs(tPacket - tCurr);
    errNext = fabs(tPacket - tNext);

    /* Searches the position that corresponds to the packet. */
    while (errNext < errCurr) {
      tCurr = tNext;
      mainToOriginCurr = mainToOriginNext;
      dateCurr++;
      errCurr = errNext;
      posFileReadResult = velodynePosFileRead(posFileId, &tNext,
                                              &mainToOriginNext);
      if (posFileReadResult == VELODYNE_READER_ERR_END_POS_FILE) {
        fprintf(stderr, "velodyne: end of position file\n");
        SDI_F->running = 0;
        pthread_mutex_unlock(&task.mtx);
        return NULL;
      } else if (posFileReadResult
                 == VELODYNE_READER_ERR_CANNOT_READ_POS_FILE) {
        fprintf(stderr, "velodyne: cannot read position file\n");
        SDI_F->running = 0;
        pthread_mutex_unlock(&task.mtx);
        return NULL;
      }
      errNext = fabs(tPacket - tNext);
    }

    /* Puts position in the buffer. */
    task.buffer.packets[task.buffer.write].position.date = dateCurr;
    task.buffer.packets[task.buffer.write].position.sensorToMain.euler
        = SDI_F->sensorToMain;
    task.buffer.packets[task.buffer.write].position.mainToOrigin.euler
        = mainToOriginCurr;

    /* Updates buffer's write pointer. */
    task.buffer.write = (task.buffer.write + 1) % task.buffer.nbPackets;

    /* Post-reading operations. */
    if (task.buffer.write == task.buffer.read)
      task.buffer.dataOverwriting = true;
    pthread_mutex_unlock(&task.mtx);
  }
  return NULL;
}

velodyneReaderResult tdImageReadVelodyne(velodyne3DImage* image, char* fileName)
{
  FILE* file;
  char aux[VELODYNE_MAX_STRLEN]; /* Auxiliary variable for fscanf. */
  int height, width, maxScanWidth;
  int i, j;
  VELODYNE_CARTESIAN_POINT cartPt;
  int result;
  velodyne3DPoint* tdPt;

  /* Opens the file and reads the format line. We check again if the file was
     correctly open, but this time we skip the check on fscanf. We suppose that,
     if the format was correctly read by the caller function, then there's no
     reason why it shouldn't be like this again. */
  file = fopen(fileName, "r");
  if (file == NULL)
    return VELODYNE_READER_ERR_CANNOT_OPEN_3D_IMAGE_FILE;
  result = fscanf(file, "%s", aux);

  /* Reads the header. */
  result = fscanf(file, "%d %d %d\n", &height, &width, &maxScanWidth);
  if (result != 3)
    return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
  if (image->height != height || image->maxScanWidth != maxScanWidth)
    return VELODYNE_READER_ERR_BAD_3D_IMAGE_SIZE;
  image->width = width;

  /* Reads the points. */
  for (i = 0; i < image->height; i++) {
    for (j = 0; j < image->width; j++) {
      result = fread(&cartPt, sizeof(VELODYNE_CARTESIAN_POINT), 1, file);
      if (result != 1)
        return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
      tdPt = velodyne3DImageAt(image, i, j);
      tdPt->coordinates[0] = cartPt.x;
      tdPt->coordinates[1] = cartPt.y;
      tdPt->coordinates[2] = cartPt.z;
      tdPt->intensity = (short) cartPt.intensity;
      tdPt->status = (velodyne3DPointStatus) cartPt.status;
    }
  }

  /* Closes the file. */
  fclose(file);
  return VELODYNE_READER_OK;
}

velodyneReaderResult tdImageReadIm3D(velodyne3DImage* image, char* fileName) {
  DATA_IM3D im3D;
  int result;
  int i, j;
  DATA_PT3D* pt3D;
  velodyne3DPoint* tdPt;

  /* Reads the 3D image. Concerning the error handling, it should be noted that
     more than one images3d error is mapped to the same error, namely
     VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE. */
  result = read_data_im3d_binaire(&im3D, fileName);
  if (result != ERR_IM3D_OK) {
    empty_data_im3d(&im3D);
    switch (result) {
    case ERR_IM3D_OPEN_ASCII_FILE:
      return VELODYNE_READER_ERR_CANNOT_OPEN_3D_IMAGE_FILE;
      break;
    /* The following case should never happen, because it was tested by the
       caller function. It's here to make the code clearer. */
    case ERR_IM3D_FORMAT_ASCII_FILE:
      return VELODYNE_READER_ERR_BAD_3D_IMAGE_FILE_FORMAT;
      break;
    case ERR_IM3D_READ_HEADER:
      return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
      break;
    case ERR_IM3D_TAILLE:
      return VELODYNE_READER_ERR_BAD_3D_IMAGE_SIZE;
      break;
    case ERR_IM3D_MALLOC_POINTS:
      return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
      break;
    case ERR_IM3D_READ_ASCII_FILE:
      return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
      break;
    default:
      return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
      break;
    }
  }

  /* Sets the width, which due to lack of information is set to the same value
     as the maximum scan width. */
  if (image->height != im3D.header.nblig
      || image->maxScanWidth != im3D.header.nbcol) {
    empty_data_im3d(&im3D);
    return VELODYNE_READER_ERR_BAD_3D_IMAGE_SIZE;
  }
  image->width = image->maxScanWidth;

  /* Copies the points. */
  for (i = 0; i < image->height; i++) {
    for (j = 0; j < image->width; j++) {
      pt3D = data_im3d_pt(&im3D, i, j);
      if (pt3D->state == GOOD_PT3D) {
        tdPt = velodyne3DImageAt(image, i, j);
        tdPt->status = VELODYNE_GOOD_3DPOINT;
        tdPt->coordinates[0] = pt3D->coord_1;
        tdPt->coordinates[1] = pt3D->coord_2;
        tdPt->coordinates[2] = pt3D->coord_3;
      }
    }
  }
  empty_data_im3d(&im3D);
  return VELODYNE_READER_OK;
}

velodyneReaderResult velodyne3DImageRead(velodyne3DImage* image, char* fileName)
{
  FILE* file;
  char format[VELODYNE_MAX_STRLEN];
  int fscanfResult;
  velodyneReaderResult result;

  /* Opens the file to check the format. */
  file = fopen(fileName, "r");
  if (file == NULL)
    return VELODYNE_READER_ERR_CANNOT_OPEN_3D_IMAGE_FILE;

  /* Reads the format, closes the file and calls the appropriate function. */
  fscanfResult = fscanf(file, "%s", format);
  if (fscanfResult != 1)
    return VELODYNE_READER_ERR_CANNOT_READ_3D_IMAGE_FILE;
  fclose(file);
  if (strcmp(format, VELODYNE_FILE_FORMAT) == 0)
    return tdImageReadVelodyne(image, fileName);
  else if (strcmp(format, FF_DATA_IM3D_BINARY) == 0)
    return tdImageReadIm3D(image, fileName);
  else
    return VELODYNE_READER_ERR_BAD_3D_IMAGE_FILE_FORMAT;
}
